"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import asyncio
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import aiohttp

from .errors import AuthenticationError, ClientError, RequestError
from .events import (IncomingWebSocketMessage, NodeConnectedEvent,
                     NodeDisconnectedEvent, NodeReadyEvent, PlayerUpdateEvent,
                     TrackEndEvent, TrackExceptionEvent, TrackStartEvent,
                     TrackStuckEvent, WebSocketClosedEvent)
from .server import EndReason, Severity
from .stats import Stats

if TYPE_CHECKING:
    from .abc import BasePlayer
    from .client import Client
    from .node import Node

_log = logging.getLogger(__name__)
CLOSE_TYPES = (
    aiohttp.WSMsgType.CLOSE,
    aiohttp.WSMsgType.CLOSING,
    aiohttp.WSMsgType.CLOSED
)
MESSAGE_QUEUE_MAX_SIZE = 25
LAVALINK_API_VERSION = 'v4'


class Transport:
    """ The class responsible for handling connections to a Lavalink server. """
    __slots__ = ('client', '_node', '_session', '_ws', '_message_queue', 'trace_requests',
                 '_host', '_port', '_password', '_ssl', 'session_id', '_destroyed')

    def __init__(self, node, host: str, port: int, password: str, ssl: bool, session_id: Optional[str], connect: bool = True):
        self.client: 'Client' = node.client
        self._node: 'Node' = node

        self._session: aiohttp.ClientSession = self.client._session
        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self._message_queue = []
        self.trace_requests = False

        self._host: str = host
        self._port: int = port
        self._password: str = password
        self._ssl: bool = ssl

        self.session_id: Optional[str] = session_id
        self._destroyed: bool = False

        if connect:
            self.connect()

    @property
    def ws_connected(self):
        """ Returns whether the websocket is connected to Lavalink. """
        return self._ws is not None and not self._ws.closed

    @property
    def http_uri(self) -> str:
        """ Returns a 'base' URI pointing to the node's address and port, also factoring in SSL. """
        return f'{"https" if self._ssl else "http"}://{self._host}:{self._port}'

    async def close(self, code=aiohttp.WSCloseCode.OK):
        """|coro|

        Shuts down the websocket connection if there is one.
        """
        if self._ws:
            await self._ws.close(code=code)
            self._ws = None

    def connect(self) -> asyncio.Task:
        """ Attempts to establish a connection to Lavalink. """
        loop = asyncio.get_event_loop()
        return loop.create_task(self._connect())

    async def destroy(self):
        """|coro|

        Closes the WebSocket gracefully, and stops any further reconnecting.
        Useful when needing to remove a node.
        """
        self._destroyed = True
        await self.close()

    async def _connect(self):
        if self._destroyed:
            raise IOError('Cannot instantiate any connections with a closed session!')

        if self._ws:
            await self.close()

        headers = {
            'Authorization': self._password,
            'User-Id': str(self.client._user_id),
            'Client-Name': f'Lavalink.py/{__import__("lavalink").__version__}'
        }

        if self.session_id is not None:
            headers['Session-Id'] = self.session_id

        _log.info('[Node:%s] Establishing WebSocket connection to Lavalink...', self._node.name)

        protocol = 'wss' if self._ssl else 'ws'
        attempt = 0

        while not self.ws_connected and not self._destroyed:
            attempt += 1
            try:
                self._ws = await self._session.ws_connect(f'{protocol}://{self._host}:{self._port}/{LAVALINK_API_VERSION}/websocket',
                                                          headers=headers,
                                                          heartbeat=60)
            except (aiohttp.ClientConnectorError, aiohttp.WSServerHandshakeError, aiohttp.ServerDisconnectedError) as error:
                if isinstance(error, aiohttp.ClientConnectorError):
                    _log.warning('[Node:%s] Invalid response received; is the server running on the correct port?',
                                 self._node.name)
                elif isinstance(error, aiohttp.WSServerHandshakeError):
                    if error.status in (401, 403):  # Special handling for 401/403 (Unauthorized/Forbidden).
                        _log.warning('[Node:%s] Authentication failed while trying to establish a connection to the node.',
                                     self._node.name)
                        # We shouldn't try to establish any more connections as correcting this particular error
                        # would require the cog to be reloaded (or the bot to be rebooted), so further attempts
                        # would be futile, and a waste of resources.
                    else:
                        _log.warning('[Node:%s] Received code \'%d\' (expected \'101\'). Check your server\'s ports and try again.',
                                     self._node.name, error.status)

                    return
                else:
                    _log.exception('[Node:%s] An unknown error occurred whilst trying to establish a connection to Lavalink', self._node.name)

                backoff = min(10 * attempt, 60)
                await asyncio.sleep(backoff)
            else:
                _log.info('[Node:%s] WebSocket connection established', self._node.name)
                await self.client._dispatch_event(NodeConnectedEvent(self._node))

                if self._message_queue:
                    for message in self._message_queue:
                        await self._send(**message)

                    self._message_queue.clear()

                attempt = 0
                await self._listen()

    async def _listen(self):
        """ Listens for websocket messages. """
        close_code: Optional[aiohttp.WSCloseCode] = None
        close_reason: Optional[str] = 'Improper websocket closure'

        assert self._ws is not None

        async for msg in self._ws:
            _log.debug('[Node:%s] Received WebSocket message: %s', self._node.name, msg.data)

            if msg.type == aiohttp.WSMsgType.TEXT:
                try:
                    await self._handle_message(msg.json())
                except Exception:  # pylint: disable=W0718
                    _log.exception('[Node:%s] Unexpected error occurred whilst processing websocket message', self._node.name)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                exc = self._ws.exception()
                _log.error('[Node:%s] Exception in WebSocket!', self._node.name, exc_info=exc)
                close_code = aiohttp.WSCloseCode.INTERNAL_ERROR
                close_reason = 'WebSocket error'
                break
            elif msg.type in CLOSE_TYPES:
                _log.debug('[Node:%s] Received close frame with code %d.', self._node.name, msg.data)
                close_code = msg.data
                close_reason = msg.extra
                break

        ws_close_code = self._ws.close_code

        if close_code is None and ws_close_code is not None:
            close_code = aiohttp.WSCloseCode(ws_close_code)

        await self.close(close_code or aiohttp.WSCloseCode.ABNORMAL_CLOSURE)
        await self._websocket_closed(close_code, close_reason)

    async def _websocket_closed(self, code: Optional[int] = None, reason: Optional[str] = None):
        """
        Handles when the websocket is closed.

        Parameters
        ----------
        code: Optional[:class:`int`]
            The response code.
        reason: Optional[:class:`str`]
            Reason why the websocket was closed. Defaults to ``None``.
        """
        _log.warning('[Node:%s] WebSocket disconnected with the following: code=%s reason=%s', self._node.name, code, reason)
        self._ws = None
        await self._node.manager._handle_node_disconnect(self._node)
        await self.client._dispatch_event(NodeDisconnectedEvent(self._node, code, reason))

    async def _handle_message(self, data: Union[Dict[Any, Any], List[Any]]):
        """
        Handles the response from the websocket.

        Parameters
        ----------
        data: Union[Dict[Any, Any], List[Any]]
            The payload received from the Lavalink server.
        """
        if self.client.has_listeners(IncomingWebSocketMessage):
            await self.client._dispatch_event(IncomingWebSocketMessage(data.copy(), self._node))

        if not isinstance(data, dict) or 'op' not in data:
            return

        op = data['op']  # pylint: disable=C0103

        if op == 'ready':
            self.session_id = data['sessionId']
            await self._node.manager._handle_node_ready(self._node)
            await self.client._dispatch_event(NodeReadyEvent(self._node, data['sessionId'], data['resumed']))
        elif op == 'playerUpdate':
            guild_id = int(data['guildId'])
            player: 'BasePlayer' = self.client.player_manager.get(guild_id)  # type: ignore

            if not player:
                _log.debug('[Node:%s] Received playerUpdate for non-existent player! GuildId: %d', self._node.name, guild_id)
                return

            if player.node != self._node:
                _log.debug('[Node:%s] Received playerUpdate for a player that doesn\'t belong to this node (player is moving?) GuildId: %d',
                           self._node.name, guild_id)
                return

            state = data['state']
            await player.update_state(state)
            await self.client._dispatch_event(PlayerUpdateEvent(player, state))
        elif op == 'stats':
            self._node.stats = Stats(self._node, data)
        elif op == 'event':
            await self._handle_event(data)
        else:
            _log.warning('[Node:%s] Received unknown op: %s', self._node.name, op)

    async def _handle_event(self, data: dict):
        """
        Handles the event from Lavalink.

        Parameters
        ----------
        data: :class:`dict`
            The data given from Lavalink.
        """
        player: 'BasePlayer' = self.client.player_manager.get(int(data['guildId']))  # type: ignore
        event_type = data['type']

        if not player:
            if event_type not in ('TrackEndEvent', 'WebSocketClosedEvent'):  # Player was most likely destroyed if it's any of these.
                _log.warning('[Node:%s] Received event type %s for non-existent player! GuildId: %s', self._node.name, event_type, data['guildId'])

            return

        event = None

        if event_type == 'TrackStartEvent':  # Always fired after track end event (for previous track), and before any track exception/stuck events.
            if player._next is not None:
                player.current = player._next
                player._next = None

            assert player.current is not None
            event = TrackStartEvent(player, player.current)
        elif event_type == 'TrackEndEvent':
            end_reason = EndReason.from_str(data['reason'])
            event = TrackEndEvent(player, player.current, end_reason)
        elif event_type == 'TrackExceptionEvent':
            exception = data['exception']
            message = exception['message']
            severity = Severity.from_str(exception['severity'])
            cause = exception['cause']

            assert player.current is not None
            event = TrackExceptionEvent(player, player.current, message, severity, cause)
        elif event_type == 'TrackStuckEvent':
            assert player.current is not None
            event = TrackStuckEvent(player, player.current, data['thresholdMs'])
        elif event_type == 'WebSocketClosedEvent':
            event = WebSocketClosedEvent(player, data['code'], data['reason'], data['byRemote'])
        else:
            _log.warning('[Node:%s] Unknown event received of type \'%s\'', self._node.name, event_type)
            return

        await self.client._dispatch_event(event)

        if player:
            try:
                await player.handle_event(event)
            except:  # noqa: E722 pylint: disable=bare-except
                _log.exception('Player %d encountered an error whilst handling event %s', player.guild_id, type(event).__name__)

    async def _send(self, **data):
        """
        Sends a payload to Lavalink.

        Parameters
        ----------
        data: :class:`dict`
            The data sent to Lavalink.
        """
        if not self.ws_connected:
            _log.debug('[Node:%s] WebSocket not ready; queued outgoing payload.', self._node.name)

            if len(self._message_queue) >= MESSAGE_QUEUE_MAX_SIZE:
                _log.warning('[Node:%s] WebSocket message queue is currently at capacity, discarding payload.', self._node.name)
            else:
                self._message_queue.append(data)

            return

        _log.debug('[Node:%s] Sending payload %s', self._node.name, str(data))
        assert self._ws is not None  # This should always pass as self.ws_connected returns False if the ws does not exist.

        try:
            await self._ws.send_json(data)
        except ConnectionResetError:
            _log.warning('[Node:%s] Failed to send payload due to connection reset!', self._node.name)

    async def _request(self, method: str, path: str, to=None, trace: bool = False, versioned: bool = True, **kwargs):  # pylint: disable=C0103
        if self._destroyed:
            raise IOError('Cannot instantiate any connections with a closed session!')

        if trace is True or self.trace_requests is True:
            kwargs['params'] = {**kwargs.get('params', {}), 'trace': 'true'}

        if versioned:
            request_url = f'{self.http_uri}/{LAVALINK_API_VERSION}/{path}'
        else:
            request_url = f'{self.http_uri}/{path}'

        _log.debug('[Node:%s] Sending request to Lavalink with the following parameters: method=%s, url=%s, params=%s, json=%s',
                   self._node.name, method, request_url, kwargs.get('params', {}), kwargs.get('json', {}))

        try:
            async with self._session.request(method=method, url=request_url,
                                             headers={'Authorization': self._password}, **kwargs) as res:
                if res.status in (401, 403):
                    raise AuthenticationError

                if res.status == 200:
                    if to is str:
                        return await res.text()

                    json = await res.json()
                    return json if to is None else to.from_dict(json)

                if res.status == 204:
                    return True

                raise RequestError('An invalid response was received from the node.',
                                   status=res.status, response=await res.json(), params=kwargs.get('params', {}))
        except (AuthenticationError, RequestError):
            raise  # Pass the caught errors back to the caller in their 'original' form.
        except Exception as original:  # It's not pretty but aiohttp doesn't specify what exceptions can be thrown.
            raise ClientError from original



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import struct
from base64 import b64decode
from io import BytesIO
from typing import Optional

from .utfm_codec import read_utfm


class DataReader:
    def __init__(self, base64_str: str):
        self._buf = BytesIO(b64decode(base64_str))

    @property
    def remaining(self) -> int:
        """ The amount of bytes left to be read. """
        return self._buf.getbuffer().nbytes - self._buf.tell()

    def _read(self, count: int):
        return self._buf.read(count)

    def read_byte(self) -> bytes:
        """
        Reads a single byte from the stream.

        Returns
        -------
        :class:`bytes`
        """
        return self._read(1)

    def read_boolean(self) -> bool:
        """
        Reads a bool from the stream.

        Returns
        -------
        :class:`bool`
        """
        result, = struct.unpack('B', self.read_byte())
        return result != 0

    def read_unsigned_short(self) -> int:
        """
        Reads an unsigned short from the stream.

        Returns
        -------
        :class:`int`
        """
        result, = struct.unpack('>H', self._read(2))
        return result

    def read_int(self) -> int:
        """
        Reads an int from the stream.

        Returns
        -------
        :class:`int`
        """
        result, = struct.unpack('>i', self._read(4))
        return result

    def read_long(self) -> int:
        """
        Reads a long from the stream.

        Returns
        -------
        :class:`int`
        """
        result, = struct.unpack('>Q', self._read(8))
        return result

    def read_nullable_utf(self, utfm: bool = False) -> Optional[str]:
        """
        .. _modified UTF: https://en.wikipedia.org/wiki/UTF-8#Modified_UTF-8

        Reads an optional UTF string from the stream.

        Internally, this just reads a bool and then a string if the bool is ``True``.

        Parameters
        ----------
        utfm: :class:`bool`
            Whether to read the string as `modified UTF`_.

        Returns
        -------
        Optional[:class:`str`]
        """
        exists = self.read_boolean()

        if not exists:
            return None

        return self.read_utfm() if utfm else self.read_utf().decode()

    def read_utf(self) -> bytes:
        """
        Reads a UTF string from the stream.

        Returns
        -------
        :class:`bytes`
        """
        text_length = self.read_unsigned_short()
        return self._read(text_length)

    def read_utfm(self) -> str:
        """
        .. _modified UTF: https://en.wikipedia.org/wiki/UTF-8#Modified_UTF-8

        Reads a UTF string from the stream.

        This method is different to :func:`read_utf` as it accounts for
        different encoding methods utilised by Java's streams, which uses `modified UTF`_
        for character encoding.

        Returns
        -------
        :class:`str`
        """
        text_length = self.read_unsigned_short()
        utf_string = self._read(text_length)
        return read_utfm(text_length, utf_string)


class DataWriter:
    def __init__(self):
        self._buf = BytesIO()

    def _write(self, data):
        self._buf.write(data)

    def write_byte(self, byte):
        """
        Writes a single byte to the stream.

        Parameters
        ----------
        byte: Any
            This can be anything ``BytesIO.write()`` accepts.
        """
        self._buf.write(byte)

    def write_boolean(self, boolean: bool):
        """
        Writes a bool to the stream.

        Parameters
        ----------
        boolean: :class:`bool`
            The bool to write.
        """
        enc = struct.pack('B', 1 if boolean else 0)
        self.write_byte(enc)

    def write_unsigned_short(self, short: int):
        """
        Writes an unsigned short to the stream.

        Parameters
        ----------
        short: :class:`int`
            The unsigned short to write.
        """
        enc = struct.pack('>H', short)
        self._write(enc)

    def write_int(self, integer: int):
        """
        Writes an int to the stream.

        Parameters
        ----------
        integer: :class:`int`
            The integer to write.
        """
        enc = struct.pack('>i', integer)
        self._write(enc)

    def write_long(self, long_value: int):
        """
        Writes a long to the stream.

        Parameters
        ----------
        long_value: :class:`int`
            The long to write.
        """
        enc = struct.pack('>Q', long_value)
        self._write(enc)

    def write_nullable_utf(self, utf_string: Optional[str]):
        """
        Writes an optional string to the stream.

        Parameters
        ----------
        utf_string: Optional[:class:`str`]
            The optional string to write.
        """
        self.write_boolean(bool(utf_string))

        if utf_string:
            self.write_utf(utf_string)

    def write_utf(self, utf_string: str):
        """
        Writes a utf string to the stream.

        Parameters
        ----------
        utf_string: :class:`str`
            The string to write.
        """
        utf = utf_string.encode('utf8')
        byte_len = len(utf)

        if byte_len > 65535:
            raise OverflowError('UTF string may not exceed 65535 bytes!')

        self.write_unsigned_short(byte_len)
        self._write(utf)

    def finish(self) -> bytes:
        """
        Finalizes the stream by writing the necessary flags, byte length etc.

        Returns
        ----------
        :class:`bytes`
            The finalized stream.
        """
        with BytesIO() as track_buf:
            byte_len = self._buf.getbuffer().nbytes
            flags = byte_len | (1 << 30)
            enc_flags = struct.pack('>i', flags)
            track_buf.write(enc_flags)

            self._buf.seek(0)
            track_buf.write(self._buf.read())
            self._buf.close()

            track_buf.seek(0)
            return track_buf.read()



# flake8: noqa

__title__ = 'Lavalink'
__author__ = 'Devoxin'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017-present Devoxin'
__version__ = '5.4.0'


from typing import Type

from .abc import BasePlayer, DeferredAudioTrack, Source
from .client import Client
from .dataio import DataReader, DataWriter
from .errors import (AuthenticationError, ClientError, InvalidTrack, LoadError,
                     RequestError)
from .events import (Event, IncomingWebSocketMessage, NodeChangedEvent,
                     NodeConnectedEvent, NodeDisconnectedEvent, NodeReadyEvent,
                     PlayerErrorEvent, PlayerUpdateEvent, QueueEndEvent,
                     TrackEndEvent, TrackExceptionEvent, TrackLoadFailedEvent,
                     TrackStartEvent, TrackStuckEvent, WebSocketClosedEvent)
from .filters import (ChannelMix, Distortion, Equalizer, Filter, Karaoke,
                      LowPass, Rotation, Timescale, Tremolo, Vibrato, Volume)
from .node import Node
from .nodemanager import NodeManager
from .player import DefaultPlayer
from .playermanager import PlayerManager
from .server import (AudioTrack, EndReason, LoadResult, LoadResultError,
                     LoadType, PlaylistInfo, Plugin, Severity)
from .source_decoders import DEFAULT_DECODER_MAPPING
from .stats import Penalty, Stats
from .utils import (decode_track, encode_track, format_time, parse_time,
                    timestamp_to_millis)


def listener(*events: Type[Event]):
    """
    Marks this function as an event listener for Lavalink.py.
    This **must** be used on class methods, and you must ensure that you register
    decorated methods by using :func:`Client.add_event_hooks`.

    Example:

        .. code:: python

            @listener()
            async def on_lavalink_event(self, event):  # Event can be ANY Lavalink event
                ...

            @listener(TrackStartEvent)
            async def on_track_start(self, event: TrackStartEvent):
                ...

    Note
    ----
    Track event dispatch order is not guaranteed!
    For example, this means you could receive a :class:`TrackStartEvent` before you receive a
    :class:`TrackEndEvent` when executing operations such as ``skip()``.

    Parameters
    ----------
    events: :class:`Event`
        The events to listen for. Leave this empty to listen for all events.
    """
    def wrapper(func):
        setattr(func, '_lavalink_events', events)
        return func
    return wrapper



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from abc import ABC, abstractmethod
from typing import (TYPE_CHECKING, Any, Dict, Generic, List, Optional, TypeVar,
                    Union)

from .common import MISSING
from .errors import InvalidTrack, LoadError
from .events import Event, TrackLoadFailedEvent
from .server import AudioTrack

if TYPE_CHECKING:
    from .client import Client
    from .node import Node
    from .server import LoadResult

_log = logging.getLogger(__name__)

FilterValueT = TypeVar('FilterValueT', Dict[str, Any], List[float], List[int], float)


class BasePlayer(ABC):
    """
    Represents the BasePlayer all players must be inherited from.

    Attributes
    ----------
    client: :class:`Client`
        The Lavalink client instance.
    guild_id: :class:`int`
        The guild id of the player.
    node: :class:`Node`
        The node that the player is connected to.
    channel_id: Optional[:class:`int`]
        The ID of the voice channel the player is connected to.
        This could be ``None`` if the player isn't connected.
    current: Optional[:class:`AudioTrack`]
        The currently playing track.
    """
    def __init__(self, guild_id: int, node: 'Node'):
        self.client: 'Client' = node.manager.client
        self.guild_id: int = guild_id
        self.node: 'Node' = node
        self.channel_id: Optional[int] = None
        self.current: Optional[AudioTrack] = None

        self._next: Optional[AudioTrack] = None
        self._internal_id: str = str(guild_id)
        self._original_node: Optional['Node'] = None  # This is used internally for failover.
        self._voice_state = {}

    @abstractmethod
    async def handle_event(self, event: Event):
        """|coro|

        Handles an :class:`Event` received directly from the websocket.

        Parameters
        ----------
        event: :class:`Event`
            The event that will be handled.
        """
        raise NotImplementedError

    @abstractmethod
    async def update_state(self, state: Dict[str, Any]):
        """|coro|

        .. _state object: https://lavalink.dev/api/websocket#player-state

        Updates this player's state with the `state object`_ received from the server.

        Parameters
        ----------
        state: Dict[:class:`str`, Any]
            The player state.
        """
        raise NotImplementedError

    async def play_track(self,
                         track: Union[AudioTrack, 'DeferredAudioTrack'],
                         start_time: int = MISSING,
                         end_time: int = MISSING,
                         no_replace: bool = MISSING,
                         volume: int = MISSING,
                         pause: bool = MISSING,
                         **kwargs) -> Optional[Dict[str, Any]]:
        """|coro|

        .. _player object: https://lavalink.dev/api/rest.html#Player

        Plays the given track.

        Warning
        -------
        Multiple calls to this method within a short timeframe could cause issues with the player's
        internal state, which can cause errors when processing a :class:`TrackStartEvent`.

        Parameters
        ----------
        track: Union[:class:`AudioTrack`, :class:`DeferredAudioTrack`]
            The track to play.
        start_time: :class:`int`
            The number of milliseconds to offset the track by.
            If left unspecified or ``None`` is provided, the track will start from the beginning.
        end_time: :class:`int`
            The position at which the track should stop playing.
            This is an absolute position, so if you want the track to stop at 1 minute, you would pass 60000.
            The default behaviour is to play until no more data is received from the remote server.
            If left unspecified or ``None`` is provided, the default behaviour is exhibited.
        no_replace: :class:`bool`
            If set to true, operation will be ignored if a track is already playing or paused.
            The default behaviour is to always replace.
            If left unspecified or ``None`` is provided, the default behaviour is exhibited.
        volume: :class:`int`
            The initial volume to set. This is useful for changing the volume between tracks etc.
            If left unspecified or ``None`` is provided, the volume will remain at its current setting.
        pause: :class:`bool`
            Whether to immediately pause the track after loading it.
            The default behaviour is to never pause.
            If left unspecified or ``None`` is provided, the default behaviour is exhibited.
        **kwargs: Any
            The kwargs to use when playing. You can specify any extra parameters that may be
            used by plugins, which offer extra features not supported out-of-the-box by Lavalink.py.

        Returns
        -------
        Optional[Dict[:class:`str`, Any]]
            The updated `player object`_, or ``None`` if a request wasn't made due to an empty payload.
        """
        if track is MISSING or not isinstance(track, AudioTrack):
            raise ValueError('track must be an instance of an AudioTrack!')

        options = kwargs

        if start_time is not MISSING:
            if not isinstance(start_time, int) or 0 > start_time:
                raise ValueError('start_time must be an int with a value equal to, or greater than 0')

            options['position'] = start_time

        if end_time is not MISSING:
            if not isinstance(end_time, int) or 1 > end_time:
                raise ValueError('end_time must be an int with a value equal to, or greater than 1')

            options['end_time'] = end_time

        if no_replace is not MISSING:
            if not isinstance(no_replace, bool):
                raise TypeError('no_replace must be a bool')

            options['no_replace'] = no_replace

        if volume is not MISSING:
            if not isinstance(volume, int):
                raise TypeError('volume must be an int')

            self.volume = max(min(volume, 1000), 0)
            options['volume'] = self.volume

        if pause is not MISSING:
            if not isinstance(pause, bool):
                raise TypeError('pause must be a bool')

            options['paused'] = pause

        playable_track = track.track

        if playable_track is None:
            if not isinstance(track, DeferredAudioTrack):
                raise InvalidTrack('Cannot play the AudioTrack as \'track\' is None, and it is not a DeferredAudioTrack!')

            try:
                playable_track = await track.load(self.client)
            except LoadError as load_error:
                await self.client._dispatch_event(TrackLoadFailedEvent(self, track, load_error))
                return

        if playable_track is None:  # This should only fire when a DeferredAudioTrack fails to yield a base64 track string.
            await self.client._dispatch_event(TrackLoadFailedEvent(self, track, None))  # type: ignore
            return

        self._next = track
        return await self.node.update_player(guild_id=self._internal_id, encoded_track=playable_track, **options)

    def cleanup(self):
        pass

    async def destroy(self):
        """|coro|

        Destroys the current player instance.

        Shortcut for :func:`PlayerManager.destroy`.
        """
        await self.client.player_manager.destroy(self.guild_id)

    async def _voice_server_update(self, data):
        self._voice_state.update(endpoint=data['endpoint'], token=data['token'])

        if 'sessionId' not in self._voice_state:  # We should've received session_id from a VOICE_STATE_UPDATE before receiving a VOICE_SERVER_UPDATE.
            _log.warning('[Player:%s] Missing sessionId, is the client User ID correct?', self.guild_id)

        await self._dispatch_voice_update()

    async def _voice_state_update(self, data):
        raw_channel_id = data['channel_id']
        self.channel_id = int(raw_channel_id) if raw_channel_id else None

        if not self.channel_id:  # We're disconnecting
            self._voice_state.clear()
            return

        if data['session_id'] != self._voice_state.get('sessionId'):
            self._voice_state.update(sessionId=data['session_id'])

            await self._dispatch_voice_update()

    async def _dispatch_voice_update(self):
        if {'sessionId', 'endpoint', 'token'} == self._voice_state.keys():
            await self.node.update_player(guild_id=self._internal_id, voice_state=self._voice_state)

    @abstractmethod
    async def node_unavailable(self):
        """|coro|

        Called when a player's node becomes unavailable.
        Useful for changing player state before it's moved to another node.
        """
        raise NotImplementedError

    @abstractmethod
    async def change_node(self, node: 'Node'):
        """|coro|

        Called when a node change is requested for the current player instance.

        Parameters
        ----------
        node: :class:`Node`
            The new node to switch to.
        """
        raise NotImplementedError


class DeferredAudioTrack(ABC, AudioTrack):
    """
    Similar to an :class:`AudioTrack`, however this track only stores metadata up until it's
    played, at which time :func:`load` is called to retrieve a base64 string which is then used for playing.

    Note
    ----
    For implementation: The ``track`` field need not be populated as this is done later via
    the :func:`load` method. You can optionally set ``self.track`` to the result of :func:`load`
    during implementation, as a means of caching the base64 string to avoid fetching it again later.
    This should serve the purpose of speeding up subsequent play calls in the event of repeat being enabled,
    for example.
    """
    @abstractmethod
    async def load(self, client: 'Client') -> Optional[str]:
        """|coro|

        Retrieves a base64 string that's playable by Lavalink.
        For example, you can use this method to search Lavalink for an identical track from other sources,
        which you can then use the base64 string of to play the track on Lavalink.

        Parameters
        ----------
        client: :class:`Client`
            This will be an instance of the Lavalink client 'linked' to this track.

        Returns
        -------
        Optional[:class:`str`]
            A Lavalink-compatible base64-encoded string containing track metadata.
            If a track string cannot be returned, you may return ``None`` or throw a :class:`LoadError`.
        """
        raise NotImplementedError


class Source(ABC):
    def __init__(self, name: str):
        self.name: str = name

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.name == other.name

        return False

    def __hash__(self):
        return hash(self.name)

    @abstractmethod
    async def load_item(self, client: 'Client', query: str) -> Optional['LoadResult']:
        """|coro|

        Loads a track with the given query.

        Parameters
        ----------
        client: :class:`Client`
            The Lavalink client. This could be useful for performing a Lavalink search
            for an identical track from other sources, if needed.
        query: :class:`str`
            The search query that was provided.

        Returns
        -------
        Optional[:class:`LoadResult`]
            A LoadResult, or ``None`` if there were no matches for the provided query.
        """
        raise NotImplementedError

    def __repr__(self):
        return f'<Source name={self.name}>'


class Filter(ABC, Generic[FilterValueT]):
    """
    A class representing a Lavalink audio filter.

    Parameters
    ----------
    values: Union[Dict[str, Any], List[Union[float, int]], float]
        The values for this filter.
    plugin_filter: :class:`bool`
        Whether this filter is part of a Lavalink plugin. Typically, this will be ``True``
        when creating your own filters.

    Attributes
    ----------
    values: Union[Dict[str, Any], List[Union[float, int]], float]
        The values for this filter.
    plugin_filter: :class:`bool`
        Whether this filter is part of a Lavalink plugin.
    """
    def __init__(self, values: FilterValueT, plugin_filter: bool = False):
        self.values: FilterValueT = values
        self.plugin_filter: bool = plugin_filter

    @abstractmethod
    def update(self, **kwargs):
        """ Updates the internal values to match those provided. """
        raise NotImplementedError

    @abstractmethod
    def serialize(self) -> Dict[str, FilterValueT]:
        """
        Transforms the internal values into a dict matching the structure Lavalink expects.

        Example:

            .. code:: python

                return {"yourCustomFilter": {"gain": 5}}
        """
        raise NotImplementedError



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from random import randrange
from time import time
from typing import (TYPE_CHECKING, Dict, List, Optional, Type,  # Literal
                    TypeVar, Union)

from .abc import BasePlayer, DeferredAudioTrack
from .common import MISSING
from .errors import ClientError, RequestError
from .events import (NodeChangedEvent, PlayerErrorEvent, QueueEndEvent,
                     TrackEndEvent, TrackStuckEvent)
from .filters import Filter
from .server import AudioTrack

if TYPE_CHECKING:
    from .node import Node

_log = logging.getLogger(__name__)

FilterT = TypeVar('FilterT', bound=Filter)


class DefaultPlayer(BasePlayer):
    """
    The player that Lavalink.py uses by default.

    This should be sufficient for most use-cases.

    Attributes
    ----------
    LOOP_NONE: :class:`int`
        Class attribute. Disables looping entirely.
    LOOP_SINGLE: :class:`int`
        Class attribute. Enables looping for a single (usually currently playing) track only.
    LOOP_QUEUE: :class:`int`
        Class attribute. Enables looping for the entire queue. When a track finishes playing, it'll be added to the end of the queue.

    guild_id: :class:`int`
        The guild id of the player.
    node: :class:`Node`
        The node that the player is connected to.
    paused: :class:`bool`
        Whether or not a player is paused.
    position_timestamp: :class:`int`
        Returns the track's elapsed playback time as an epoch timestamp.
    volume: :class:`int`
        The volume at which the player is playing at.
    shuffle: :class:`bool`
        Whether or not to mix the queue up in a random playing order.
    loop: Literal[0, 1, 2]
        Whether loop is enabled, and the type of looping.
        This is an integer as loop supports multiple states.

        0 = Loop off.

        1 = Loop track.

        2 = Loop queue.

        Example
        -------
        .. code:: python

            if player.loop == player.LOOP_NONE:
                await ctx.send('Not looping.')
            elif player.loop == player.LOOP_SINGLE:
                await ctx.send(f'{player.current.title} is looping.')
            elif player.loop == player.LOOP_QUEUE:
                await ctx.send('This queue never ends!')
    filters: Dict[:class:`str`, :class:`Filter`]
        A mapping of str to :class:`Filter`, representing currently active filters.
    queue: List[:class:`AudioTrack`]
        A list of AudioTracks to play.
    current: Optional[:class:`AudioTrack`]
        The track that is playing currently, if any.
    """
    LOOP_NONE: int = 0
    LOOP_SINGLE: int = 1
    LOOP_QUEUE: int = 2

    def __init__(self, guild_id: int, node: 'Node'):
        super().__init__(guild_id, node)

        self._user_data = {}

        self.paused: bool = False
        self._internal_pause: bool = False  # Toggled when player's node becomes unavailable, primarily used for track position tracking.
        self._last_update: int = 0
        self._last_position: int = 0
        self.position_timestamp: int = 0
        self.volume: int = 100
        self.shuffle: bool = False
        self.loop: int = 0  # 0 = off, 1 = single track, 2 = queue
        self.filters: Dict[str, Filter] = {}
        self.queue: List[AudioTrack] = []

    @property
    def is_playing(self) -> bool:
        """ Returns the player's track state. """
        return self.is_connected and self.current is not None

    @property
    def is_connected(self) -> bool:
        """ Returns whether the player is connected to a voicechannel or not. """
        return self.channel_id is not None

    @property
    def position(self) -> int:
        """ Returns the track's elapsed playback time in milliseconds, adjusted for Lavalink stat interval. """
        if not self.is_playing:
            return 0

        assert self.current is not None

        if self.paused or self._internal_pause:
            return min(self._last_position, self.current.duration)

        difference = int(time() * 1000) - self._last_update
        return min(self._last_position + difference, self.current.duration)

    def store(self, key: object, value: object):
        """
        Stores custom user data.

        Parameters
        ----------
        key: :class:`object`
            The key of the object to store.
        value: :class:`object`
            The object to associate with the key.
        """
        self._user_data.update({key: value})

    def fetch(self, key: object, default=None):
        """
        Retrieves the related value from the stored user data.

        Parameters
        ----------
        key: :class:`object`
            The key to fetch.
        default: Optional[:class:`any`]
            The object that should be returned if the key doesn't exist. Defaults to ``None``.

        Returns
        -------
        Optional[:class:`any`]
        """
        return self._user_data.get(key, default)

    def delete(self, key: object):
        """
        Removes an item from the the stored user data.

        Parameters
        ----------
        key: :class:`object`
            The key to delete.

        Raises
        ------
        :class:`KeyError`
            If the key doesn't exist.
        """
        try:
            del self._user_data[key]
        except KeyError:
            pass

    def add(self, track: Union[AudioTrack, 'DeferredAudioTrack', Dict[str, Union[Optional[str], bool, int]]],
            requester: int = 0, index: Optional[int] = None):
        """
        Adds a track to the queue.

        Parameters
        ----------
        track: Union[:class:`AudioTrack`, :class:`DeferredAudioTrack`, Dict[str, Union[Optional[str], bool, int]]]
            The track to add. Accepts either an AudioTrack or
            a dict representing a track returned from Lavalink.
        requester: :class:`int`
            The ID of the user who requested the track.
        index: Optional[:class:`int`]
            The index at which to add the track.
            If index is left unspecified, the default behaviour is to append the track. Defaults to ``None``.
        """
        track = AudioTrack(track, requester) if isinstance(track, dict) else track

        if requester != 0:
            track.requester = requester

        if index is None:
            self.queue.append(track)
        else:
            self.queue.insert(index, track)

    async def play(self,
                   track: Optional[Union[AudioTrack, 'DeferredAudioTrack', Dict[str, Union[Optional[str], bool, int]]]] = None,
                   start_time: int = MISSING,
                   end_time: int = MISSING,
                   no_replace: bool = MISSING,
                   volume: int = MISSING,
                   pause: bool = MISSING,
                   **kwargs):
        """|coro|

        Plays the given track.

        This method differs from :func:`BasePlayer.play_track` in that it contains additional logic
        to handle certain attributes, such as ``loop``, ``shuffle``, and loading a base64 string from :class:`DeferredAudioTrack`.

        Warning
        -------
        Multiple calls to this method within a short timeframe could cause issues with the player's
        internal state, which can cause errors when processing a :class:`TrackStartEvent`.

        Parameters
        ----------
        track: Optional[Union[:class:`DeferredAudioTrack`, :class:`AudioTrack`, Dict[str, Union[Optional[str], bool, int]]]]
            The track to play. If left unspecified, this will default to the first track in the queue. Defaults to ``None``
            which plays the next song in queue. Accepts either an AudioTrack or a dict representing a track
            returned from Lavalink.
        start_time: :class:`int`
            The number of milliseconds to offset the track by.
            If left unspecified, the track will start from the beginning.
        end_time: :class:`int`
            The position at which the track should stop playing.
            This is an absolute position, so if you want the track to stop at 1 minute, you would pass 60000.
            If left unspecified, the track will play through to the end.
        no_replace: :class:`bool`
            If set to true, operation will be ignored if the player already has a current track.
            If left unspecified, the currently playing track will always be replaced.
        volume: :class:`int`
            The initial volume to set. This is useful for changing the volume between tracks etc.
            If left unspecified, the volume will remain at its current setting.
        pause: :class:`bool`
            Whether to immediately pause the track after loading it. Defaults to ``False``.
        **kwargs: Any
            The kwargs to use when playing. You can specify any extra parameters that may be
            used by plugins, which offer extra features not supported out-of-the-box by Lavalink.py.

        Raises
        ------
        :class:`ValueError`
            If invalid values were provided for ``start_time`` or ``end_time``.
        :class:`TypeError`
            If wrong types were provided for ``no_replace``, ``volume`` or ``pause``.
        """
        if isinstance(no_replace, bool) and no_replace and self.is_playing:
            return

        if track is not None and isinstance(track, dict):
            track = AudioTrack(track, 0)

        if self.loop > 0 and self.current:
            if self.loop == 1:
                if track is not None:
                    self.queue.insert(0, self.current)
                else:
                    track = self.current
            elif self.loop == 2:
                self.queue.append(self.current)

        self._last_position = 0
        self.position_timestamp = 0

        if pause is not MISSING and isinstance(pause, bool):
            self.paused = pause

        if not track:
            if not self.queue:
                await self.stop()  # Also sets current to None.
                await self.client._dispatch_event(QueueEndEvent(self))
                return

            pop_at = randrange(len(self.queue)) if self.shuffle else 0
            track = self.queue.pop(pop_at)

        if start_time is not MISSING:
            if not isinstance(start_time, int) or not 0 <= start_time < track.duration:
                raise ValueError('start_time must be an int with a value equal to, or greater than 0, and less than the track duration')

            self._last_position = start_time

        if end_time is not MISSING:
            if not isinstance(end_time, int) or not 1 <= end_time <= track.duration:
                raise ValueError('end_time must be an int with a value equal to, or greater than 1, and less than, or equal to the track duration')

        response = await self.play_track(track, start_time, end_time, no_replace,
                                         volume, pause, **kwargs)

        if response is not None:
            self.paused = response['paused']

    async def stop(self):
        """|coro|

        Stops the player.
        """
        await self.node.update_player(guild_id=self._internal_id, encoded_track=None)
        self.current = None

    async def skip(self):
        """|coro|

        Plays the next track in the queue, if any.

        Warning
        -------
        Multiple calls to this method within a short timeframe could cause issues with the player's
        internal state, which can cause errors when processing a :class:`TrackStartEvent`.
        """
        await self.play()

    def set_loop(self, loop: int):
        """
        Sets whether the player loops between a single track, queue or none.

        0 = off, 1 = single track, 2 = queue.

        Parameters
        ----------
        loop: Literal[0, 1, 2]
            The loop setting. 0 = off, 1 = single track, 2 = queue.
        """
        if not 0 <= loop <= 2:
            raise ValueError('Loop must be 0, 1 or 2.')

        self.loop = loop

    def set_shuffle(self, shuffle: bool):
        """
        Sets the player's shuffle state.

        Parameters
        ----------
        shuffle: :class:`bool`
            Whether to shuffle the player or not.
        """
        self.shuffle = shuffle

    async def set_pause(self, pause: bool):
        """|coro|

        Sets the player's paused state.

        Parameters
        ----------
        pause: :class:`bool`
            Whether to pause the player or not.
        """
        await self.node.update_player(guild_id=self._internal_id, paused=pause)
        self.paused = pause

    async def set_volume(self, vol: int):
        """|coro|

        Sets the player's volume

        Note
        ----
        A limit of 1000 is imposed by Lavalink.

        Parameters
        ----------
        vol: :class:`int`
            The new volume level.
        """
        vol = max(min(vol, 1000), 0)
        await self.node.update_player(guild_id=self._internal_id, volume=vol)
        self.volume = vol

    async def seek(self, position: int):
        """|coro|

        Seeks to a given position in the track.

        Parameters
        ----------
        position: :class:`int`
            The new position to seek to in milliseconds.
        """
        if not isinstance(position, int):
            raise ValueError('position must be an int!')

        await self.node.update_player(guild_id=self._internal_id, position=position)

    async def set_filters(self, *filters: Filter):
        """|coro|

        This sets multiple filters at once.

        Applies the corresponding filters within Lavalink.
        This will overwrite any identical filters that are already applied.

        Parameters
        ----------
        *filters: :class:`Filter`
            The filters to apply.

        Raises
        ------
        :class:`TypeError`
            If any of the provided filters is not of type :class:`Filter`.
        """
        for _filter in filters:
            if not isinstance(_filter, Filter):
                raise TypeError(f'Expected object of type Filter, not {type(_filter).__name__}')

            filter_name = type(_filter).__name__.lower()
            self.filters[filter_name] = _filter

        await self._apply_filters()

    async def set_filter(self, _filter: Filter):
        """|coro|

        Applies the corresponding filter within Lavalink.
        This will overwrite the filter if it's already applied.

        Example
        -------
        .. code:: python

            equalizer = Equalizer()
            equalizer.update(bands=[(0, 0.2), (1, 0.3), (2, 0.17)])
            player.set_filter(equalizer)

        Parameters
        ----------
        _filter: :class:`Filter`
            The filter instance to set.

        Raises
        ------
        :class:`TypeError`
            If the provided ``_filter`` is not of type :class:`Filter`.
        """
        if not isinstance(_filter, Filter):
            raise TypeError(f'Expected object of type Filter, not {type(_filter).__name__}')

        filter_name = type(_filter).__name__.lower()
        self.filters[filter_name] = _filter
        await self._apply_filters()

    async def update_filter(self, _filter: Type[FilterT], **kwargs):
        """|coro|

        Updates a filter using the upsert method;
        if the filter exists within the player, its values will be updated;
        if the filter does not exist, it will be created with the provided values.

        This will not overwrite any values that have not been provided.

        Example
        -------
        .. code :: python

            player.update_filter(Timescale, speed=1.5)
            # This means that, if the Timescale filter is already applied
            # and it already has set values of "speed=1, pitch=1.2", pitch will remain
            # the same, however speed will be changed to 1.5 so the result is
            # "speed=1.5, pitch=1.2"

        Parameters
        ----------
        _filter: Type[:class:`Filter`]
            The filter class (**not** an instance of, see above example) to upsert.
        **kwargs: Any
            The kwargs to pass to the filter.

        Raises
        ------
        :class:`TypeError`
            If the provided ``_filter`` is not of type :class:`Filter`.
        """
        if isinstance(_filter, Filter):
            raise TypeError(f'Expected class of type Filter, not an instance of {type(_filter).__name__}')

        if not issubclass(_filter, Filter):
            raise TypeError(f'Expected subclass of type Filter, not {_filter.__name__}')

        filter_name = _filter.__name__.lower()

        filter_instance = self.filters.get(filter_name, _filter())  # type: ignore
        filter_instance.update(**kwargs)
        self.filters[filter_name] = filter_instance
        await self._apply_filters()

    def get_filter(self, _filter: Union[Type[FilterT], str]):
        """
        Returns the corresponding filter, if it's enabled.

        Example
        -------
        .. code:: python

            from lavalink.filters import Timescale
            timescale = player.get_filter(Timescale)
            # or
            timescale = player.get_filter('timescale')

        Parameters
        ----------
        _filter: Union[Type[:class:`Filter`], :class:`str`]
            The filter name, or filter class (**not** an instance of, see above example), to get.

        Returns
        -------
        Optional[:class:`Filter`]
        """
        if isinstance(_filter, str):
            filter_name = _filter
        elif isinstance(_filter, Filter):  # User passed an instance of.
            filter_name = type(_filter).__name__
        else:
            if not issubclass(_filter, Filter):
                raise TypeError(f'Expected subclass of type Filter, not {_filter.__name__}')

            filter_name = _filter.__name__

        return self.filters.get(filter_name.lower(), None)

    async def remove_filters(self, *filters: Union[Type[FilterT], str]):
        """|coro|

        Removes multiple filters from the player, undoing any effects applied to the audio.
        This is similar to :func:`remove_filter` but instead allows you to remove multiple filters with one call.

        Parameters
        ----------
        filters: Union[Type[:class:`Filter`], :class:`str`]
            The filters to remove. Can be filter name, or filter class (**not** an instance of).
        """
        for fltr in filters:
            await self.remove_filter(fltr)

    async def remove_filter(self, _filter: Union[Type[FilterT], str]):
        """|coro|

        Removes a filter from the player, undoing any effects applied to the audio.

        Example
        -------
        .. code:: python

            player.remove_filter(Timescale)
            # or
            player.remove_filter('timescale')

        Parameters
        ----------
        _filter: Union[Type[:class:`Filter`], :class:`str`]
            The filter name, or filter class (**not** an instance of, see above example), to remove.
        """
        if isinstance(_filter, str):
            filter_name = _filter
        elif isinstance(_filter, Filter):  # User passed an instance of.
            filter_name = type(_filter).__name__
        else:
            if not issubclass(_filter, Filter):
                raise TypeError(f'Expected subclass of type Filter, not {_filter.__name__}')

            filter_name = _filter.__name__

        fn_lowered = filter_name.lower()

        if fn_lowered in self.filters:
            self.filters.pop(fn_lowered)
            await self._apply_filters()

    async def clear_filters(self):
        """|coro|

        Clears all currently-enabled filters.
        """
        self.filters.clear()
        await self._apply_filters()

    async def _apply_filters(self):
        await self.node.update_player(guild_id=self._internal_id, filters=list(self.filters.values()))

    async def handle_event(self, event):
        """
        Handles the given event as necessary.

        Parameters
        ----------
        event: :class:`Event`
            The event to handle.
        """
        # A track throws loadFailed when it fails to provide any audio before throwing an exception.
        # A TrackStuckEvent is not proceeded by a TrackEndEvent. In theory, you could ignore a TrackStuckEvent
        # and hope that a track will eventually play, however, it's unlikely.

        if isinstance(event, TrackStuckEvent) or isinstance(event, TrackEndEvent) and event.reason.may_start_next():
            try:
                await self.play()
            except RequestError as error:
                await self.client._dispatch_event(PlayerErrorEvent(self, error))
                _log.exception('[DefaultPlayer:%d] Encountered a request error whilst starting a new track.', self.guild_id)

    async def update_state(self, state: dict):
        """
        Updates the position of the player.

        Parameters
        ----------
        state: :class:`dict`
            The state that is given to update.
        """
        self._last_update = int(time() * 1000)
        self._last_position = state.get('position', 0)
        self.position_timestamp = state.get('time', 0)

    async def node_unavailable(self):
        """|coro|

        Called when a player's node becomes unavailable.
        Useful for changing player state before it's moved to another node.
        """
        self._last_position = self.position
        self._internal_pause = True

    async def change_node(self, node: 'Node'):
        """|coro|

        Changes the player's node

        Parameters
        ----------
        node: :class:`Node`
            The node the player is changed to.
        """
        old_node = self.node
        self.node = node

        last_position = self.position

        try:
            await self.node.destroy_player(self._internal_id)
        except (ClientError, RequestError):
            pass

        if self._voice_state:
            await self._dispatch_voice_update()

        if self.current:
            playable_track = self.current.track

            if isinstance(self.current, DeferredAudioTrack) and playable_track is None:
                playable_track = await self.current.load(self.client)

            self._last_position = last_position  # Ensure that _last_position is correctly set, in case a node sends us bad data.

            await self.node.update_player(guild_id=self._internal_id, encoded_track=playable_track, position=last_position,
                                          paused=self.paused, volume=self.volume)
            self._last_update = int(time() * 1000)

        self._internal_pause = False

        if self.filters:
            await self._apply_filters()

        await self.client._dispatch_event(NodeChangedEvent(self, old_node, node))

    def __repr__(self):
        return f'<DefaultPlayer volume={self.volume} current={self.current}>'



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple

from .errors import ClientError
from .node import Node

if TYPE_CHECKING:
    from .client import Client

_log = logging.getLogger(__name__)
DEFAULT_REGIONS = {
    'asia': ('hongkong', 'singapore', 'sydney', 'japan', 'southafrica', 'india'),
    'eu': ('rotterdam', 'russia'),
    'us': ('us-central', 'us-east', 'us-south', 'us-west', 'brazil')
}


class NodeManager:
    """Represents the node manager that contains all lavalink nodes.

    len(x):
        Returns the total number of nodes.
    iter(x):
        Returns an iterator of all the stored nodes.

    Attributes
    ----------
    client: :class:`Client`
        The Lavalink client.
    nodes: List[:class:`Node`]
        Cache of all the nodes that Lavalink has created.
    regions: Dict[str, Tuple[str]]
        A mapping of continent -> Discord RTC regions.
    """
    __slots__ = ('_player_queue', '_connect_back', 'client', 'nodes', 'regions')

    def __init__(self, client, regions: Optional[Dict[str, Tuple[str]]], connect_back: bool):
        self._player_queue = []
        self._connect_back: bool = connect_back
        self.client: 'Client' = client
        self.nodes: List[Node] = []
        self.regions: Dict[str, Tuple[str]] = regions or DEFAULT_REGIONS

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterator[Node]:
        yield from self.nodes

    @property
    def available_nodes(self) -> List[Node]:
        """
        Returns a list of available nodes.

        .. deprecated:: 5.0.0
            As of Lavalink server 4.0.0, a WebSocket connection is no longer required to operate a
            node. As a result, this property is no longer considered useful as all nodes are considered
            available.
        """
        return [n for n in self.nodes if n.available]

    def add_node(self, host: str, port: int, password: str, region: str, name: Optional[str] = None,
                 ssl: bool = False, session_id: Optional[str] = None, connect: bool = True) -> Node:
        """
        Adds a node to this node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        region: :class:`str`
            The region to assign this node to.
        name: Optional[:class:`str`]
            An identifier for the node that will show in logs. Defaults to ``None``.
        ssl: :class:`bool`
            Whether to use SSL for the node. SSL will use ``wss`` and ``https``, instead of ``ws`` and ``http``,
            respectively. Your node should support SSL if you intend to enable this, either via reverse proxy or
            other methods. Only enable this if you know what you're doing.
        session_id: Optional[:class:`str`]
            The ID of the session to resume. Defaults to ``None``.
            Only specify this if you have the ID of the session you want to resume.
        connect: :class:`bool`
            Whether to immediately connect to the node after creating it.
            If ``False``, you must call :func:`Node.connect` if you require WebSocket functionality.

        Returns
        -------
        :class:`Node`
            The created Node instance.
        """
        node = Node(self, host, port, password, region, name, ssl,
                    session_id, connect)
        self.nodes.append(node)
        return node

    def remove_node(self, node: Node):
        """
        Removes a node.

        Make sure you have called :func:`Node.destroy` to close any resources used by this Node.

        .. deprecated:: 5.2.0
            To be consistent with function naming, this method has been deprecated in favour of
            :func:`remove`.

        Parameters
        ----------
        node: :class:`Node`
            The node to remove from the list.
        """
        self.nodes.remove(node)

    def remove(self, node: Node):
        """
        Removes a node.

        Make sure you have called :func:`Node.destroy` to close any resources used by this Node.

        Parameters
        ----------
        node: :class:`Node`
            The node to remove from the list.
        """
        self.nodes.remove(node)

    def get_nodes_by_region(self, region_key: str) -> List[Node]:
        """
        Get a list of nodes by their region.
        This does not account for node availability, so the nodes returned
        could be either available or unavailable.

        Parameters
        ----------
        region_key: :class:`str`
            The region key. If you haven't specified your own regions, this will be
            one of ``asia``, ``eu`` or ``us``, otherwise, it'll be one of the keys
            within the dict you provided.

        Returns
        -------
        List[:class:`Node`]
            A list of nodes. Could be empty if no nodes were found with the specified region key.
        """
        return [n for n in self.nodes if n.region == region_key]

    def get_region(self, endpoint: str) -> Optional[str]:
        """
        Returns a Lavalink.py-friendly region from a Discord voice server address.

        Parameters
        ----------
        endpoint: :class:`str`
            The address of the Discord voice server.

        Returns
        -------
        Optional[:class:`str`]
        """
        if not endpoint:
            return None

        endpoint = endpoint.replace('vip-', '')

        for key in self.regions:
            if not any(n.region == key for n in self.available_nodes):
                continue

            if endpoint.startswith(self.regions[key]):
                return key

        return None

    def find_ideal_node(self, region: Optional[str] = None, exclude: Optional[List[Node]] = None) -> Optional[Node]:
        """
        Finds the best (least used) node in the given region, if applicable.

        Parameters
        ----------
        region: Optional[:class:`str`]
            The region to find a node in. Defaults to ``None``.
        exclude: Optional[List[:class:`Node`]]
            A list of nodes to exclude from the choice.

        Returns
        -------
        Optional[:class:`Node`]
        """
        exclusions = exclude or []
        nodes = None

        if region:
            nodes = [n for n in self.available_nodes if n.region == region and n not in exclusions]

        if not nodes:  # If there are no regional nodes available, or a region wasn't specified.
            nodes = [n for n in self.available_nodes if n not in exclusions]

        if not nodes:
            return None

        best_node = min(nodes, key=lambda node: node.penalty)
        return best_node

    async def _handle_node_ready(self, node: Node):
        for player in self._player_queue:
            await player.change_node(node)
            original_node_name = player._original_node.name if player._original_node else '[no node]'
            _log.debug('Moved player %d from node \'%s\' to node \'%s\'', player.guild_id, original_node_name, node.name)

        if self._connect_back:
            for player in node._original_players:
                await player.change_node(node)
                player._original_node = None

        self._player_queue.clear()

    async def _handle_node_disconnect(self, node: Node):
        """|coro|

        Called when a node is disconnected from Lavalink.

        Parameters
        ----------
        node: :class:`Node`
            The node that has just connected.
        """
        for player in node.players:
            try:
                await player.node_unavailable()
            except:  # noqa: E722 pylint: disable=bare-except
                _log.exception('An error occurred whilst calling player.node_unavailable()')

        best_node = self.find_ideal_node(node.region, exclude=[node])  # Don't use the node these players are moving from.

        if not best_node:
            self._player_queue.extend(node.players)
            _log.warning('Unable to move players, no available nodes! Waiting for a node to become available.')
            return

        # TODO: This may need reinvestigating to make it more robust with the lack of WS requirement.
        # i.e. we need a way to determine whether nodes are "reachable".
        for player in node.players:
            try:
                await player.change_node(best_node)

                if self._connect_back:
                    player._original_node = node
            except ClientError:
                _log.error('Failed to move player %d from node \'%s\' to new node \'%s\'', player.guild_id, node.name, best_node.name)
